@@ -1,14 +1,13 @@
 """Overview.
 =============
 
-This module simplifies the management of function default keyword argument 
-values.
-
-:py:mod:`funconf` introduces two decorators. The first function
-:py:func:`wraps_var_kwargs` makes it possible to dynamically define the
-default kwargs for a function that takes a variable number of kwargs.  The
-second function :py:func:`lazy_string_cast` casts input parameters passed in
-based on the default values of the kwargs it was constructed with. 
+To simplify the management of function default keyword argument values
+:py:mod:`funconf` introduces two new decorators. The first function
+:py:func:`wraps_kwargs` makes it trivial to dynamically define the default
+kwargs for a function.  The second function :py:func:`lazy_string_cast`
+automatically casts *basestring* values to the type of the keyword default
+values found in the function it is wrapping and the type of the values found in
+the *key:value* object passed into its constructor. 
 
 For configuration, :py:mod:`funconf` borrows from concepts discussed in
 Python's core library *ConfigParser*.  A configuration consists of sections
@@ -63,7 +62,7 @@ A configuration object
 
 The :py:class:`Config` class is the root class container for all of the
 configuration sections.  It implements the *MutableMapping* type which allows
-it to seamlessly integrate into the :py:func:`wraps_var_kwargs` function. 
+it to seamlessly integrate into the :py:func:`wraps_kwargs` function. 
 
 As a dictionary ``dict(config)`` the :py:class:`Config` object looks like::
 
@@ -74,8 +73,8 @@ As a dictionary ``dict(config)`` the :py:class:`Config` object looks like::
     
 Notice how the configuration *section:options* have been concatenated. This
 facilitates the simple wrapping of an entire configuration file into the kwargs
-of a function.  The following example will print out an equivalent dictionary
-to the printout in the previous example. ::  
+of a function.  The following example will print an equivalent dictionary to
+the printout in the previous example. ::  
 
     @config
     def myfunc(**kwargs):
@@ -93,7 +92,7 @@ To access sections and option values the following pattern holds true::
 
 A section is represented by the :py:class:`ConfigSection` object. This object
 implements the *MutableMapping* type in the same way the :py:class:`Config`
-object is implemented and is compatible with the :py:func:`wraps_var_kwargs`
+object is implemented and is compatible with the :py:func:`wraps_kwargs`
 function.  The :py:class:`ConfigSection` represented ``str(config.bread)``
 looks like::
 
@@ -110,6 +109,7 @@ used as a decorator to a function too. Here is an example::
 
 """
 import functools
+import inspect
 from collections import MutableMapping
 import shlex
 from distutils.util import strtobool
@@ -117,48 +117,77 @@ try:
     from inspect import signature, Signature, Parameter
 except ImportError:
     from funcsigs import signature, Signature, Parameter 
+try:
+    u = unicode
+except NameError:
+    u = lambda x: x
+from io import StringIO
 import yaml
 
 
-def wraps_var_kwargs(kwarg_defaults):
-    """Decorate a function to define set of *key:value* pairs and hide the
-    variable keyword argument parameter. 
+def wraps_kwargs(default_kwargs):
+    """Decorate a function to define and extend its default keyword argument
+    values.
         
     The following example will redefine myfunc to have default kwargs of a=4
     and b=2:: 
 
         mydict = {a=4, b=2}
-        @wraps_var_kwargs(mydict)
+        @wraps_kwargs(mydict)
         def myfunc(**k):
             pass
 
-    The *kwarg_defaults* object needs to satisfy the *MutableMapping* interface
+    The *default_kwargs* object needs to satisfy the *MutableMapping* interface
     definition. When the wrapped function is called the kwargs passed in are
-    used to update the *kwarg_defaults* object. The *kwarg_defaults* object is
-    then passed into the wrapped function i.e. ``wrapped(**kwarg_defaults)``. 
-    
-    :param kwarg_defaults: kwargs to be fix into the wrapped function.
-    :type kwarg_defaults: mutable mapping
+    used to update the *default_kwargs* object. The *default_kwargs* object is
+    then passed into the wrapped function i.e. ``wrapped(**default_kwargs)``. 
+
+
+    :param default_kwargs: kwargs to be fix into the wrapped function.
+    :type default_kwargs: mutable mapping
     :rtype: decorated function.
     """
     def decorator(func):
-        kwarg_defaults_set = set(kwarg_defaults)
+        default_kwargs_set = set(default_kwargs)
+        func_defaults_set = set()
+        override_defaults_set = set()
+        has_var_keyword = False
         def wrapper(**kwargs):
-            diff = set(kwargs) - kwarg_defaults_set 
-            if diff:
-                msg = "%s() got an unexpected keyword argument '%s'" % \
-                        (func.func_name, "', '".join(diff)) 
-                raise TypeError(msg)
-            kwarg_defaults.update(kwargs)
-            return func(**kwarg_defaults)
+            kwargs_set = set(kwargs)
+            # Update default_kwargs with new keyword values.
+            for k in kwargs_set.intersection(default_kwargs_set):
+                default_kwargs[k] = kwargs[k]
+            # Override func's default keyword values.
+            for k in override_defaults_set.difference(kwargs_set):
+                kwargs[k] = default_kwargs[k]
+            if has_var_keyword:
+                # Add default_kwargs keyword values not defined in kwargs.
+                for k in default_kwargs_set.difference(kwargs_set):
+                    kwargs[k] = default_kwargs[k]
+            else:
+                # Remove kwargs that func doesn't have defined.
+                for k in kwargs_set.difference(func_defaults_set):
+                    kwargs.pop(k)
+            return func(**kwargs)
+
         funcsig = signature(func)
-        for param in funcsig.parameters.values():
-            if param.kind != param.VAR_KEYWORD:
-                msg = "Can only wrap variable length keyword arguments"
+        parameters = []
+        for arg, param in funcsig.parameters.items():
+            if param.default != param.empty:
+                func_defaults_set.add(arg)
+                if arg in default_kwargs_set:
+                    override_defaults_set.add(arg)
+                else:
+                    param = Parameter(arg, Parameter.KEYWORD_ONLY,
+                                           default=param.default)
+                    parameters.append(param)
+            elif param.kind == param.VAR_KEYWORD:
+                has_var_keyword = True
+            else:
+                msg = "%s parameters are not supported" % param.kind 
                 raise ValueError(msg)
         functools.update_wrapper(wrapper, func)
-        parameters = []
-        for k, v in kwarg_defaults.items():
+        for k, v in default_kwargs.items():
             param = Parameter(k, Parameter.KEYWORD_ONLY, default=v)
             parameters.append(param)
         sig = Signature(parameters=parameters)
@@ -167,7 +196,7 @@ def wraps_var_kwargs(kwarg_defaults):
     return decorator
 
 
-def lazy_string_cast(model_kwargs):
+def lazy_string_cast(model_kwargs={}):
     """Type cast string input values if they differ from the type of the
     default value found in *model_kwargs*.
     
@@ -189,19 +218,19 @@ def lazy_string_cast(model_kwargs):
     
     This example demonstrates how :py:func:`lazy_string_cast` can be applied::
         
-        default = dict(a=4, b=[4, 2, 55])
-
-        @lazy_string_cast(default)
+        @lazy_string_cast()
         def main(a=4, b=[4, 2, 55]):
             pass
 
-    Or using :py:func:`lazy_string_cast` with :py:func:`wraps_var_kwargs`::
 
-        default = dict(a=4, b=[4, 2, 55])
+    Or using :py:func:`lazy_string_cast` with :py:func:`wraps_kwargs` to
+    define new keyword defaults::
 
-        @lazy_string_cast(default)
-        @wraps_var_kwargs(default)
-        def main(**k):
+        config = dict(a=6, b=[4])
+
+        @lazy_string_cast(config)
+        @wraps_kwargs()
+        def main(a=4, b=[4, 2, 55]):
             pass
 
     :param model_kwargs: kwargs to model default type values and keys from.
@@ -212,7 +241,7 @@ def lazy_string_cast(model_kwargs):
         try:
             value = vtype(value)
         except:
-            msg = "Can not convert %s='%s' to %s" % (k, value,
+            msg = "Can not convert %s='%s' to %s" % (key, value,
                     vtype)
             raise ValueError(msg)
         return value
@@ -238,17 +267,22 @@ def lazy_string_cast(model_kwargs):
             inner_cast_func = None if not v else cast_factory(k, v[0])
             return make_cast_func(cast_list, k, inner_cast_func)
         elif vtype is bool:
-            return make_cast_func(cast_type_raise, k, strtobool)
+            return make_cast_func(cast_type_raise, k, 
+                                  lambda x: bool(strtobool(x)))
         elif vtype in [int, float]:
             return make_cast_func(cast_type_raise, k, vtype)
         else:
             return make_cast_func(cast_type, k, vtype)
 
-    cast_ctrl = {}
-    for k, v in model_kwargs.items():
-        cast_ctrl[k] = (type(v), cast_factory(k, v))
-
     def decorator(func):
+        cast_ctrl = {}
+        for k, v in model_kwargs.items():
+            cast_ctrl[k] = (type(v), cast_factory(k, v))
+        funcsig = signature(func)
+        for k, param in funcsig.parameters.items():
+            if param.default != param.empty:
+                v = param.default
+                cast_ctrl[k] = (type(v), cast_factory(k, v))
         @functools.wraps(func)
         def wrapper(**kwargs):
             for k, v in kwargs.items():
@@ -258,7 +292,12 @@ def lazy_string_cast(model_kwargs):
                         kwargs[k] = cast_func(v)
             return func(**kwargs)
         return wrapper
-    return decorator
+
+    if inspect.isfunction(model_kwargs) or inspect.ismethod(model_kwargs):
+        func, model_kwargs = model_kwargs, {}
+        return decorator(func)
+    else:
+        return decorator
 
 
 class ConfigAttributeError(AttributeError): pass
@@ -273,14 +312,11 @@ class ConfigSection(MutableMapping):
     * exposes a configuration *option:value* through a standard implementation
       of the *MutableMapping* abstract type.
     * when cast to a string it outputs its state in YAML.
-    * as a decorator it utilises the :py:func:`wraps_var_kwargs` to change
+    * as a decorator it utilises the :py:func:`wraps_kwargs` to change
       the defaults of a variable kwargs function.  
     """
-
-    _dirty = None 
-    _options = None
-    _section = None
-    _reserved = None
+ 
+    __slots__ = ('_dirty', '_options', '_section', '_reserved')
 
     def __init__(self, section, options):
         """Construct a new :py:class:`ConfigSection` object.  
@@ -319,7 +355,7 @@ class ConfigSection(MutableMapping):
             return self._options[y]
 
     def __setattr__(self, x, y):
-        "Only attributes that are a reserved work can be set in this object."
+        "Only attributes that are a reserved words can be set in this object."
         if x in ConfigSection._reserved:
             super(ConfigSection, self).__setattr__(x, y)
         else:
@@ -390,7 +426,7 @@ class ConfigSection(MutableMapping):
         :rtype: wrapped function with defined kwargs defaults bound to this 
                 :py:class:`ConfigSection` object.
         """
-        return lazy_string_cast(self)(wraps_var_kwargs(self)(func)) 
+        return lazy_string_cast(self)(wraps_kwargs(self)(func)) 
 
 
 ConfigSection._reserved = set(dir(ConfigSection))
@@ -407,15 +443,13 @@ class Config(MutableMapping):
     * exposes a translation of the configuration into section_option:value
       through a standard implementation of the *MutableMapping* abstract type.
     * when cast to a string it outputs its state in YAML.
-    * as a decorator it utilises the :py:func:`wraps_var_kwargs` to change
+    * as a decorator it utilises the :py:func:`wraps_kwargs` to change
       the defaults of a variable kwargs function.  
     """
 
-    _sections = None
-    _reserved = None
-    _lookup = None
+    __slots__ = ('_sections', '_reserved', '_lookup', '_strict')
 
-    def __init__(self, filenames=[]):
+    def __init__(self, filenames=[], strict=False):
         """Construct a new Config object.  
         
         This is the root object for a function configuration set.  It is the
@@ -423,9 +457,13 @@ class Config(MutableMapping):
 
         :param filenames: YAML configuration files.
         :type filenames: list of filepaths
+        :param strict: If True, raise :py:class:`ConfigAttributeError` if a
+                       :py:class:`ConfigSection` doesn't exist.
+        :type strict: False 
         """
         self._sections = {}
         self._lookup = {}
+        self._strict = strict 
         self.read(filenames)
 
     def read(self, filenames):
@@ -460,6 +498,8 @@ class Config(MutableMapping):
         :param stream: the configuration to be loaded using ``yaml.load``.
         :type stream: stream object
         """
+        if isinstance(stream, basestring):
+            stream = StringIO(u(stream))
         config = yaml.load(stream)
         if not isinstance(config, dict):
             return
@@ -512,12 +552,15 @@ class Config(MutableMapping):
             return super(Config, self).__getattribute__(y)
         else:
             if y not in self._sections:
-                msg = "Config object has no section '%s'" % (y)
-                raise ConfigAttributeError(msg)
+                if not self._strict:
+                    self._sections[y] = ConfigSection(y, {})
+                else:
+                    msg = "Config object has no section '%s'" % (y)
+                    raise ConfigAttributeError(msg)
             return self._sections[y]
 
     def __setattr__(self, x, y):
-        "Only attributes that are a reserved work can be set in this object."
+        "Only attributes that are a reserved words can be set in this object."
         if x in Config._reserved:
             super(Config, self).__setattr__(x, y)
         else:
@@ -531,7 +574,8 @@ class Config(MutableMapping):
         return self._lookup.__iter__()
 
     def __len__(self):
-        "Return the number of options defined in this :py:class:`Config` object"
+        """Return the number of options defined in this :py:class:`Config`
+        object"""
         return len(self._lookup)
 
     def __setitem__(self, x, y):
@@ -547,7 +591,7 @@ class Config(MutableMapping):
     def __getitem__(self, y):
         "Return the option value for y where y is *section_option*."
         if y not in self._lookup:
-            raise ValueError("There is no section for '%s'" % y)
+            raise KeyError("There is no section for '%s'" % y)
         s, option = self._lookup[y]
         section = self._sections[s]
         return section[option]
@@ -575,7 +619,7 @@ class Config(MutableMapping):
         :rtype: wrapped function with defined kwargs bound to this
                 :py:class:`Config` object.
         """
-        return lazy_string_cast(self)(wraps_var_kwargs(self)(func)) 
+        return lazy_string_cast(self)(wraps_kwargs(self)(func)) 
 
 
 Config._reserved = set(dir(Config))
